use std::collections::HashMap;
use std::time::Instant;

pub fn execute_task01(content: &str) {
    let start = Instant::now();
    let sum_of_numbers = solve_01(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(25231, sum_of_numbers);
    println!("Day04 - Task01 - Duration {duration:?}- Sum of correct Numbers: {}", sum_of_numbers)
}

pub fn execute_task02(content: &str) {
    let start = Instant::now();
    let sum_of_numbers = solve_02(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(9721255, sum_of_numbers);
    println!("Day04 - Task02 - Duration {duration:?} - Sum of correct Numbers: {}", sum_of_numbers)
}

pub fn solve_01(data: &str) -> anyhow::Result<usize> {
    Ok(data
        .lines()
        .map(|data| parse_card(data))
        .filter_map(|(winning, check)| {
            let correct = winning.iter().filter(|number| check.contains(number)).count();

            if correct > 0 {
                return Some(correct);
            }
            None
        })
        .map(|num| 2_usize.pow((num-1) as u32))
        .sum())
}



fn solve_02(data: &str) -> anyhow::Result<usize> {
    let list: Vec<usize> = data
        .lines()
        .map(|data| parse_card(data))
        .map(|(winning, check)| winning.iter().filter(|number| check.contains(number)).count())
        .collect();

    let mut data: HashMap<usize, usize> = HashMap::new();

    for (index, value) in list.iter().enumerate() {
        if !data.contains_key(&index) {
            data.insert(index, 1);
        }

        for i in 1..=value.clone() as i32 {
            if list.len() <= (i as usize + index) {
                continue;
            }

            *data.entry(index + i as usize).or_insert(1) += 1 * data[&index].clone();
        }
    }

    Ok(data.values().sum())
}

fn parse_card(line: &str) -> (Vec<usize>, Vec<usize>) {
    line
        .split_once(":").unwrap().1
        .split_once("|")
        .map(|(winning, check)| (parse_number_line(winning), parse_number_line(check)))
        .unwrap()
}

fn parse_number_line(line: &str) -> Vec<usize> {
    line
        .trim()
        .split(" ")
        .filter(|value| value.len() > 0)
        .map(|value| value.parse::<usize>().unwrap())
        .collect()
}

#[test]
pub fn test_solve_01() {
    let example = r#"Card 1: 41 48 83 86 17 | 83 86  6 31 17  9 48 53
Card 2: 13 32 20 16 61 | 61 30 68 82 17 32 24 19
Card 3:  1 21 53 59 44 | 69 82 63 72 16 21 14  1
Card 4: 41 92 73 84 69 | 59 84 76 51 58  5 54 83
Card 5: 87 83 26 28 32 | 88 30 70 12 93 22 82 36
Card 6: 31 18 13 56 72 | 74 77 10 23 35 67 36 11"#;

    let result = solve_01(example).unwrap();

    assert_eq!(13, result);
}

#[test]
pub fn test_solve_02() {
    let example = r#"Card 1: 41 48 83 86 17 | 83 86  6 31 17  9 48 53
Card 2: 13 32 20 16 61 | 61 30 68 82 17 32 24 19
Card 3:  1 21 53 59 44 | 69 82 63 72 16 21 14  1
Card 4: 41 92 73 84 69 | 59 84 76 51 58  5 54 83
Card 5: 87 83 26 28 32 | 88 30 70 12 93 22 82 36
Card 6: 31 18 13 56 72 | 74 77 10 23 35 67 36 11"#;

    let result = solve_02(example).unwrap();

    assert_eq!(30, result);
}