use std::collections::{HashMap};
use std::time::Instant;
use rayon::prelude::*;
use num::Integer;

struct Node {
    left: String,
    right: String
}

pub fn execute_task01(content: &str) {
    let start = Instant::now();
    let steps = solve_01(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(22199, steps);
    println!("Day08 - Task01 - Duration: {duration:?} - Command Steps: {}", steps)
}

pub fn execute_task02(content: &str) {
    let start = Instant::now();
    let command_steps = solve_02(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(13334102464297, command_steps);
    println!("Day08 - Task02 - Duration: {duration:?} - Lot of Command Steps: {}", command_steps)
}

fn solve_01(content: &str) -> anyhow::Result<usize> {
    let mut lines = content.lines();
    let mut nodes: HashMap<String, Node> = HashMap::new();
    let cmds = lines.next().unwrap();
    lines.next().unwrap();
    lines.for_each(|line| {
        let data = line.split_once(" = ").unwrap();
        let key = data.0.to_owned();

        let node_data = data.1.split_once(", ").unwrap();
        nodes.insert(key, Node{
            left: node_data.0.replace("(", ""),
            right: node_data.1.replace(")", ""),
        });
    });

    let mut key = "AAA";
    let mut steps = 0;
    let cmds = cmds.chars();

    loop {
        for cmd in cmds.clone().into_iter() {
            if key == "ZZZ" {
                return Ok(steps)
            }

            let node = &nodes[key];
            steps += 1;
            if cmd == 'L' {
                key = node.left.as_str();
            } else {
                key = node.right.as_str();
            }
        }
    }
}

fn solve_02(content: &str) -> anyhow::Result<u64> {
    let mut lines = content.lines();
    let mut nodes: HashMap<String, Node> = HashMap::new();
    let cmds = lines.next().unwrap();
    lines.next().unwrap();
    lines.for_each(|line| {
        let data = line.split_once(" = ").unwrap();
        let key = data.0.to_owned();

        let node_data = data.1.split_once(", ").unwrap();
        nodes.insert(key, Node {
            left: node_data.0.replace("(", ""),
            right: node_data.1.replace(")", ""),
        });
    });

    let keys = nodes.keys().filter(|key| key.ends_with("A")).map(|str| str.to_owned()).collect::<Vec<String>>();
    //let end = nodes.keys().filter(|key| key.ends_with("Z")).map(|str| str.to_owned()).collect::<HashSet<String>>();
    let cmds = cmds.chars();

    let step_count: Vec<(String, u64)> = keys.par_iter()
        .map(|key| {
            let mut key = key.to_owned();
            let mut steps: u64 = 0;

            loop {
                for c in cmds.clone().into_iter() {
                    if  key.ends_with("Z") {
                        return (key, steps)
                    }

                    steps += 1;
                    let node = &nodes[&key];
                    if c == 'L' {
                        key = node.left.clone();
                    } else {
                        key = node.right.clone();
                    }
                }
            }
        }).collect();

    Ok(step_count.iter()
        .map(|(_, steps)| steps)
        .fold(1, |a, b| a.lcm(b))
        .to_owned())
}

#[test]
fn test_solve_01_1() {
    let example = r#"RL

AAA = (BBB, CCC)
BBB = (DDD, EEE)
CCC = (ZZZ, GGG)
DDD = (DDD, DDD)
EEE = (EEE, EEE)
GGG = (GGG, GGG)
ZZZ = (ZZZ, ZZZ)"#;

    let result = solve_01(example).unwrap();

    assert_eq!(2, result);
}

#[test]
fn test_solve_01_2() {
    let example = r#"LLR

AAA = (BBB, BBB)
BBB = (AAA, ZZZ)
ZZZ = (ZZZ, ZZZ)"#;

    let result = solve_01(example).unwrap();

    assert_eq!(6, result);
}

#[test]
fn test_solve_02() {
    let example = r#"LR

11A = (11B, XXX)
11B = (XXX, 11Z)
11Z = (11B, XXX)
22A = (22B, XXX)
22B = (22C, 22C)
22C = (22Z, 22Z)
22Z = (22B, 22B)
XXX = (XXX, XXX)"#;

    let result = solve_02(example).unwrap();

    assert_eq!(6, result);
}
