use std::time::Instant;

pub fn execute_task01(content: &str) {
    let start = Instant::now();
    let sum_of_b = solve_01(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(1953784198, sum_of_b);
    println!("Day09 - Task01 - Duration: {duration:?} - Sum of B: {}", sum_of_b)
}

pub fn execute_task02(content: &str) {
    let start = Instant::now();
    let command_steps = solve_02(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(957, command_steps);
    println!("Day09 - Task02 - Duration: {duration:?} - Sum of B: {}", command_steps)
}

fn solve_01(content: &str) -> anyhow::Result<i64> {
    Ok(content
        .lines()
        .map(|line| line
            .split(" ")
            .map(|num| num.parse::<i64>().unwrap())
            .collect::<Vec<i64>>())
        .map(|data_line| build_tree(data_line))
        .map(|tree| expand_tree_width_right(tree))
        .map(|tree| tree.first().unwrap().last().unwrap().to_owned())
        .sum())
}

fn solve_02(content: &str) -> anyhow::Result<i64> {
    Ok(content
        .lines()
        .map(|line| line
            .split(" ")
            .map(|num| num.parse::<i64>().unwrap())
            .collect::<Vec<i64>>())
        .map(|data_line| build_tree(data_line))
        .map(|tree| expand_tree_width_left(tree))
        .map(|tree| tree.first().unwrap().first().unwrap().to_owned())
        .sum())
}

fn build_tree(line: Vec<i64>) -> Vec<Vec<i64>> {
    let mut tree = vec![line];

    loop {
        let last_layer = tree.last().unwrap();
        if last_layer.iter().filter(|num| num != &&0).count() == 0 {
            return tree
        }

        let mut new_layer = vec![];

        for  (index, item) in last_layer.iter().enumerate() {
            if index + 1 >= last_layer.len() {
                break
            }

            new_layer.push( last_layer[index+1] - item)
        }

        tree.push(new_layer);
    }
}

fn expand_tree_width_right(tree: Vec<Vec<i64>>) -> Vec<Vec<i64>> {
    let mut tree = tree.clone();
    let tree_lines = tree.len();

    for index in 0..tree_lines {
        if index == 0 {
            tree.last_mut().unwrap().push(0);
            continue
        }

        let line = tree[tree_lines - 1 - index].clone();
        let line_before = tree[tree_lines - index].clone();

        tree
            .get_mut(tree_lines - 1 - index)
            .unwrap()
            .push(line.last().unwrap() + line_before.last().unwrap().clone())
    }

    tree
}

fn expand_tree_width_left(tree: Vec<Vec<i64>>) -> Vec<Vec<i64>> {
    let mut tree = tree.clone();
    let tree_lines = tree.len();

    for index in 0..tree_lines {
        if index == 0 {
            tree.last_mut().unwrap().insert(0, 0);
            continue
        }

        let line = tree[tree_lines - 1 - index].clone();
        let line_before = tree[tree_lines - index].clone();

        tree
            .get_mut(tree_lines - 1 - index)
            .unwrap()
            .insert(0, line.first().unwrap() - line_before.first().unwrap().clone() )
    }

    tree
}

#[test]
fn test_solve_01() {
    let example = r#"0 3 6 9 12 15
1 3 6 10 15 21
10 13 16 21 30 45"#;

    let result = solve_01(example).unwrap();

    assert_eq!(114, result);
}

#[test]
fn test_solve_02() {
    let example = r#"0 3 6 9 12 15
1 3 6 10 15 21
10 13 16 21 30 45"#;

    let result = solve_02(example).unwrap();

    assert_eq!(2, result);
}
