use std::cmp;
use std::time::Instant;
use itertools::Itertools;
use num::abs;

pub fn execute_task01(content: &str) {
    let start = Instant::now();
    let sum_of_b = solve_01(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(10165598, sum_of_b);
    println!("Day11 - Task01 - Duration: {duration:?} - Sum of closest Points: {}", sum_of_b)
}

pub fn execute_task02(content: &str) {
    let start = Instant::now();
    let command_steps = solve_02(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(678728808158, command_steps);
    println!("Day11 - Task02 - Duration: {duration:?} - Sum of closest million Points: {}", command_steps)
}


fn solve_01(content: &str) -> anyhow::Result<usize> {
    solve(content, 2)
}

fn solve_02(content: &str) -> anyhow::Result<usize> {
    solve(content, 1_000_000)
}

fn solve(content: &str, expansion: usize) -> anyhow::Result<usize> {
    let height = content.lines().count();
    let width = content.lines().next().unwrap().chars().count();

    let mut empty_y: Vec<usize> = (0..height).collect();
    let mut empty_x: Vec<usize> = (0..width).collect();

    // Find all Galaxies
    let mut galaxies = content
        .lines()
        .enumerate()
        .flat_map(|(y, line)|
            line
                .chars()
                .enumerate()
                .filter_map(|(x, c)| {
                    if c == '#' {
                        return Some((y, x));
                    }
                    None
                })
                .collect::<Vec<(usize, usize)>>()
        )
        .collect::<Vec<(usize, usize)>>();


    // TODO move this shit in the first loop but ownership and move....
    galaxies
        .iter()
        .for_each(|(y, x)| {
            empty_y.retain(|value| *value != *y);
            empty_x.retain(|value| *value != *x);
        });

    // Expand all cords
    galaxies = galaxies
        .iter()
        .map(|(y, x)| {
            (y + (empty_y.iter().take_while(|filler| filler < &y).count() * (expansion - 1)), x + (empty_x.iter().take_while(|filler| filler < &x).count() * (expansion - 1)))
        })
        .collect();

    // Calculate diff and sum them up
    Ok(galaxies.clone().iter()
        .combinations(2)
        .map(|data| (data[0], data[1]))
        .map(|(from, to)| calculate_diff(from.clone(), to.clone()))
        .sum())
}

fn calculate_diff(from: (usize, usize), to: (usize, usize)) -> usize {
    let min_y = cmp::min(from.0, to.0) as i32;
    let min_x = cmp::min(from.1, to.1) as i32;
    let max_y = cmp::max(from.0, to.0) as i32;
    let max_x = cmp::max(from.1, to.1) as i32;

    (abs(max_y - min_y) + abs(max_x - min_x)) as usize
}

#[test]
fn test_solve_01() {
    let example = r#"...#......
.......#..
#.........
..........
......#...
.#........
.........#
..........
.......#..
#...#....."#;

    let result = solve_01(example).unwrap();

    assert_eq!(374, result);
}

#[test]
fn test_solve_02() {
    let example = r#"...#......
.......#..
#.........
..........
......#...
.#........
.........#
..........
.......#..
#...#....."#;

    let result = solve(example, 10).unwrap();

    assert_eq!(1030, result);
}


#[test]
fn test_solve_02_2() {
    let example = r#"...#......
.......#..
#.........
..........
......#...
.#........
.........#
..........
.......#..
#...#....."#;

    let result = solve(example, 100).unwrap();

    assert_eq!(8410, result);
}