use std::hash::Hash;
use std::time::Instant;
use itertools::Itertools;

#[derive(Debug, Clone)]
struct Lens {
    label: String,
    hash: u32,
    value: usize,
}
pub fn execute_task01(content: &str) {
    let start = Instant::now();
    let sum_of_hash = solve_01(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(508552, sum_of_hash);
    println!("Day15 - Task01 - Duration: {duration:?} - Sum of Hash: {}", sum_of_hash)
}

pub fn execute_task02(content: &str) {
    let start = Instant::now();
    let boxes_sum = solve_02(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(265462, boxes_sum);
    println!("Day15 - Task02 - Duration: {duration:?} - Sum of Baxes: {}", boxes_sum)
}


fn solve_01(content: &str) -> anyhow::Result<u32> {
    Ok(content
        .split(",")
        .map(|part| part
            .chars()
            .fold(0, |sum, char| hash(sum + char as u32) ))
        .sum())
}



fn solve_02(content: &str) -> anyhow::Result<usize> {
    let mut boxes: Vec<Vec<Lens>> = vec![vec![]; 256];

    let _ = content
        .split(",")
        .for_each(|part| {
            if part.contains("=") {
                let (label, num) = part.split_once("=").unwrap();
                let lens = Lens{
                    label: label.to_owned(),
                    hash: hash_string(label),
                    value: num.parse().unwrap(),
                };
                let mut lable_box = &mut boxes[lens.hash as usize];

                if let Some(item) = lable_box.iter_mut().filter(|box_lens| box_lens.label == lens.label).next() { // already exists
                    item.value = lens.value
                } else {
                    lable_box.push(lens)
                }

            } else { // is dash (-)
                let label = part.replace("-", "");
                let hash = hash_string(label.as_str());
                let mut lable_box= &mut boxes[hash as usize];

                for i in 0..lable_box.len() {
                    let lens = &lable_box[i];
                    if lens.label == label {
                        lable_box.remove(i);
                        break
                    }
                }
            }
        });

    Ok(boxes
        .iter()
        .enumerate()
        .filter(|(_, list)| !list.is_empty())
        .flat_map(|(i, lenses)| lenses.iter().enumerate().map(move |(j, lens)| (i+1) * (j+1) * lens.value))
        .sum())
}

fn hash_string(str: &str) -> u32 {
    str
        .chars()
        .fold(0, |sum, char| hash(sum + char as u32) )
}

fn hash(num: u32) -> u32 {
    num * 17 % 256
}

#[test]
fn test_solve_01() {
    let example = r#"rn=1,cm-,qp=3,cm=2,qp-,pc=4,ot=9,ab=5,pc-,pc=6,ot=7"#;

    let result = solve_01(example).unwrap();

    assert_eq!(1320, result);
}

#[test]
fn test_solve_02() {
    let example = r#"rn=1,cm-,qp=3,cm=2,qp-,pc=4,ot=9,ab=5,pc-,pc=6,ot=7"#;

    let result = solve_02(example).unwrap();

    assert_eq!(145, result);
}
